;;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;; gthxml-ags-parse.lisp

;;;; Expressions to parse alternative gene structures from gthXML documents

;;;; Michael E Sparks (mespar1@iastate.edu)
;;;; Last Modified: 5 February 2007

;;;; Copyright (c) 2007 Michael E Sparks
;;;; All rights reserved.
;;;;
;;;; You are granted the rights to distribute and use this software
;;;; as governed by the terms of the Lisp Lesser General Public License
;;;; (http://opensource.franz.com/preamble.html), also known as the LLGPL.

(in-package :toolsaq)

;;; Define various AGS-specific objects and their specialized methods:

(defclass ags-exon-features ()
  ((start
    :accessor start
    :type integer
    :initform -1
    :documentation "Genomic start coordinate of AGS exon.")
   (stop
    :accessor stop
    :type integer
    :initform -1
    :documentation "Genomic stop coordinate of AGS exon.")
   (score
    :accessor score
    :type short-float
    :initarg :score
    :documentation "Score of AGS exon."))
  (:documentation "Store features of exons in an AGS."))

(defmethod print-object ((object ags-exon-features) stream)
  (print-unreadable-object (object stream :type nil :identity nil)
    (with-accessors ((start start) (stop stop) (score score)) object
		    (format stream "~&Exon~cstart: ~d~cstop: ~d~cscore: ~,3f~%"
			    #\tab start #\tab stop #\tab score))))

(defclass ags-intron-features ()
  ((donor-prob
    :accessor donor-prob
    :type short-float
    :initarg :donor-prob
    :documentation "Probability of donor site in AGS intron.")
   (acceptor-prob
    :accessor acceptor-prob
    :type short-float
    :initarg :acceptor-prob
    :documentation "Probability of acceptor site in AGS intron."))
  (:documentation "Store features of introns in an AGS."))

(defmethod print-object ((object ags-intron-features) stream)
  (print-unreadable-object (object stream :type nil :identity nil)
    (with-accessors ((dp donor-prob) (ap acceptor-prob)) object
		    (format stream "~&Intron~cdon-prob: ~,3f~cacc-prob: ~,3f~%"
			    #\tab dp #\tab ap))))

(defclass ags-features ()
  ((ags-id
    :accessor ags-id
    :type integer
    :initarg :ags-id
    :documentation "Serial number of AGS at a given PGL.")
   (gdna-id
    :accessor gdna-id
    :type string
    :initform ""
    :documentation "Genomic template identifier.")
   (exon-list
    :accessor exon-list
    :type list
    :initform '()
    :documentation "List of ags-exon-features objects,
ordered as encountered in gthXML document.")
   (intron-list
    :accessor intron-list
    :type list
    :initform '()
    :documentation "List of ags-intron-features objects,
ordered as encountered in gthXML document.")
   (support-hash
    :accessor support-hash
    :type hash-table
    :initform (make-hash-table :test #'equal)
    :documentation "Hash of functional evidence,
either DNA or protein, supporting the AGS:

key   = id string of query sequence
value = list of exon start/stop coordinates,
        which are stored as integers in an alist.
        These are the exons the query alignment
        contributed to the AGS being considered.

Note that it is assumed all strings denoting a
particular cDNA or protein sequence in a gthXML
document are case-identical."))
  (:documentation "Store relevant features of an AGS."))

(defmethod print-object ((object ags-features) stream)
  (print-unreadable-object (object stream :type nil :identity nil)
    (with-accessors ((ai ags-id)
		     (gi gdna-id)
		     (el exon-list)
		     (il intron-list)
		     (sh support-hash))
      object
      (format stream "~&AGS ~d at locus from ~a~%" ai gi)
      (labels ((print-exint-lists (elist ilist)
		 (unless (endp elist)
		   (princ (car elist) stream)
		   (if (not (endp ilist))
		       (princ (car ilist) stream))
		   (print-exint-lists (cdr elist) (cdr ilist))))
	       (report-exon-coors (alist loc-string)
		 (if (endp alist)
		     loc-string
		     (report-exon-coors
		      (cdr alist)
		      (concatenate 'string
				   loc-string " (" (caar alist)
				   " " (cdar alist) ")"))))
	       (print-support-hash (sh)
		 (maphash #'(lambda (k v)
			      (format stream "~a:~a~%"
				      k (report-exon-coors v "")))
			  sh)))
	(print-exint-lists el il)
	(format stream "Functional evidence:~%")
	(print-support-hash sh)))))

(defvar *tmp-supp-pgs-alist* '()
  "Temporarily stores PGS coordinates an alignment contributes to an AGS.")

(defvar *current-ags* nil
  "Records the AGS currently being processed, i.e., an ags-features object.")

(defvar *agss-at-pgl* nil
  "Records a list of ags-features objects from a given PGL.")

(defun print-recorded-pgls (recorded-pgls stream)
  "Print recorded PGLs to the stream indicated."
  (let ((pgl-counter 0)
	(pgl-total (list-length recorded-pgls)))
    (dolist (curr-pgl recorded-pgls)
      (format stream "~&~%*** Reporting PGL ~d of ~d ***~%"
	      (incf pgl-counter) pgl-total)
      (let ((ags-counter 0)
	    (ags-total (list-length curr-pgl)))
	(dolist (curr-ags curr-pgl)
	  (format stream "~%Reporting AGS ~d of ~d~%"
		  (incf ags-counter) ags-total)
	  (format stream "~a~%" curr-ags))))))

(defvar *recorded-pgls* nil
  "Records PGLs that satisfy the record-pgl-if predicate.")

;;; Expressions for representing the quality of an AGS

(defclass qual-ags-parms ()
  ((min-ags-per-pgl
    :accessor min-ags-per-pgl
    :type integer
    :initarg :min-ags-per-pgl
    :initform 1
    :documentation "Minimum number of AGSs at PGLs of interest.")
   (min-exon-count-per-ags
    :accessor min-exon-count-per-ags
    :type integer
    :initarg :min-exon-count-per-ags
    :initform 1
    :documentation "Minimum number of exons at AGSs of interest.")
   (use-low-supp-ct-p
    :accessor use-low-supp-ct-p
    :initform t
    :documentation "Base AGS adequacy on number of good supportive alignments?")
   (min-ags-good-supp-ct
    :accessor min-ags-good-supp-ct
    :type integer
    :initarg :min-exon-count-per-ags
    :initform 1
    :documentation "Minimum number of good supportive alignments for AGSs.")
   (min-ags-good-supp-ratio
    :accessor min-ags-good-supp-ratio
    :type short-float
    :initarg :min-ags-good-supp-ratio
    :initform 0.5
    :documentation "Minimum ratio of good:all supportive alignments for AGSs."))
  (:documentation "Parameters governing identification of interesting AGSs."))

(defmethod print-object ((object qual-ags-parms) stream)
  (print-unreadable-object (object stream :type nil :identity nil)
    (with-accessors ((mapp min-ags-per-pgl)
		     (mecpa min-exon-count-per-ags)
		     (ulscp use-low-supp-ct-p)
		     (magsc min-ags-good-supp-ct)
		     (magsr min-ags-good-supp-ratio))
      object
      (format stream "~&Min count of AGSs per PGL: ~d~%" mapp)
      (format stream "Min count of exons per AGS: ~d~%" mecpa)
      (if ulscp
	  (format stream
		  "Adequate AGSs have at least ~a good supportive alignments.~%" magsc)
	  (format stream
		  "Adequate AGSs have at least ~a good:all supportive alignments.~%" magsr)))))

(defmethod validate-parms-obj-qual ((obj qual-ags-parms))
  (with-accessors ((magspgl min-ags-per-pgl)
		   (mecags min-exon-count-per-ags)
		   (magsc min-ags-good-supp-ct)
		   (magsr min-ags-good-supp-ratio))
    obj
    (if (minusp magspgl)
	(fix-negative-value magspgl "minimum count of AGSs per PGL"))
    (if (minusp mecags)
	(fix-negative-value mecags "minimum count of exons per PGL"))
    (if (minusp magsc)
	(fix-negative-value magsc "minimum count of good alignments per AGS"))
    (if (minusp magsr)
	(fix-bad-ratio magsc "ratio of good:all alignments per AGS"))))

(defclass ags-alt-splc-parm (qual-ags-parms)
  ;; At present, this class just overrides initializations for 
  ;; slots, so the print-object method specialized on its superclass
  ;; will work for this class, also.
  ((min-ags-per-pgl
    :initform 2
    :documentation "Minimum number of AGSs at alternatively spliced PGLs.")
   (min-exon-count-per-ags
    :initform 2
    :documentation "Minimum number of exons in a spliced isoform (AGS)."))
  (:documentation "Parameters governing those AGSs that can be seen
as positively supporting alternative splicing events at genetic loci."))

(defparameter *ags-alt-splc-parms* (make-instance 'ags-alt-splc-parm)
  "This object stores parameters influencing the identification
of qualifying AGSs that support the notion of alternative splicing
events at a genetic locus.")

;;; Expressions governing when PGLs should be recorded

(defun record-pgl-if (candidate-list &key filter test)
  "First, pre-process candidate-list with filter, then
if the filtered list is still worth recording, according
to test, do so."
  (let* ((filtered-list (funcall filter candidate-list))
	 (pass-p (funcall test filtered-list)))
    (if pass-p (push filtered-list *recorded-pgls*))
    pass-p))

(defvar *query-alns-hash* nil
  "Stores alignment features of functional evidence supporting AGSs.")

;;; Build up to a filter for alternative splice detection:

(defun support-item-adequate-p (supp-id supp-exons gdna-of-ags)
  "Report whether the current unit of support for the AGS can be
found in the *query-alns-hash* object, and hence is adequate."
  (let* ((response (multiple-value-list (gethash supp-id *query-alns-hash*)))
	 (alns-entry (car response))
	 (alns-entry-present-p (cadr response)))
    (if (and alns-entry-present-p
	     ;; Let me briefly discuss the logic behind the following s-expr:
	     ;; Some alignment exists in alns-entry s.t.
	     ;; 1) gDNA of the alignment and of the AGS are the same, .AND.
	     ;; 2) For every supporting exon in the AGS' supporting exon list
	     ;;        there exists an aln-exon in aln's exons s.t.
	     ;;            the supporting AGS exon and aln-exon are equal
	     (some #'(lambda (aln)
		       (and (funcall *string-comparator* (gdna-id aln) gdna-of-ags)
			    (every #'(lambda (supp-exon)
				       (some #'(lambda (aln-exon)
						 (equal supp-exon aln-exon))
					     (gdna-exon-coors aln)))
				   supp-exons)))
		   alns-entry))
	t
	nil)))

(defun find-adequate-support (supp-hash gdna-of-ags)
  "Returns a list of strings, whose elts are id's of good AGS-supporting alignments."
  (let ((good-support '()))
    (maphash #'(lambda (key val)
		 (if (support-item-adequate-p key val gdna-of-ags)
		     (push key good-support)))
	     supp-hash)
    good-support))

(defun ags-is-adequate-p (good-supp-ct tot-supp-ct)
  "Report whether the current AGS is good enough for retention."
  (with-accessors ((ulscp use-low-supp-ct-p)
		   (magsc min-ags-good-supp-ct)
		   (magsr min-ags-good-supp-ratio))
    *ags-alt-splc-parms*
    (cond (ulscp (if (>= good-supp-ct magsc) t nil))
	  (t (if (>= (/ good-supp-ct tot-supp-ct) magsr) t nil)))))
		   
(defun retain-only-supported-agss (ags-list)
  "For each AGS entry in ags-list, this function inspects the
functional evidence supporting it, and determines whether it
is of adequate quality to be retained; the (potentially pruned)
list will be returned.  Requires that *query-alns-hash*
be set to something reasonable."
  ;; Note that, because *query-alns-hash* was set using
  ;; parse-qualified-alns-from-gthxml, all that is necessary here
  ;; is to check whether the supportive data, at the locus of
  ;; interest, exists in this object---the alignment will be of
  ;; adequate quality, already.
  (let ((loc-list '()))
    (dolist (curr-ags ags-list)
      (let ((sh (support-hash curr-ags))
	    (gdna-of-ags (gdna-id curr-ags)))
	(if (ags-is-adequate-p
	     (list-length (find-adequate-support sh gdna-of-ags))
	     (hash-table-count sh))
	    (push curr-ags loc-list))))
    (reverse loc-list)))

(defun alt-splc-detected-p (ags-list)
  "Predicate to identify PGLs harboring alternatively spliced genes."
  (with-accessors ((magspgl min-ags-per-pgl)
		   (mecags min-exon-count-per-ags))
    *ags-alt-splc-parms*
    (if (and (>= (list-length ags-list) magspgl)
	     (every #'(lambda (x)
			(>= (list-length (exon-list x))
			    mecags))
		    ags-list))
	t
	nil)))

;;; Create and set hooks for an xml-parser-state object:

(defun parse-ags-gthxml-new-element-hook (name attributes seed)
  (macrolet ((xml-elt-matches-p (query-name namespace)
	       `(and (funcall *string-comparator* ,query-name name)
		 (funcall *string-comparator* ,namespace (get-current-xmlns)))))
    (cond ((xml-elt-matches-p "GTH_output" "http://www.genomethreader.org/GTH_output/")
	   (setf *recorded-pgls* '()))
	  ((xml-elt-matches-p "predicted_gene_location" "http://www.genomethreader.org/GTH_output/PGL_module/")
	   (setf *agss-at-pgl* '()))
	  ((xml-elt-matches-p "AGS_line" "http://www.genomethreader.org/GTH_output/PGL_module/")
	   (setf *current-ags*
		 (make-instance 'ags-features
				:ags-id (parse-integer
					 (get-attr-value ':|AGS_serial| attributes)))))
	  ((xml-elt-matches-p "exon" "http://www.genomethreader.org/GTH_output/PGL_module/predicted_gene_location/AGS_information/exon-intron_info/")
	   (push (make-instance 'ags-exon-features
				:score (read-from-string
					(get-attr-value ':|e_score| attributes)))
		 (exon-list *current-ags*)))
	  ((xml-elt-matches-p "gDNA_exon_boundary" "http://www.genomethreader.org/GTH_output/PGL_module/predicted_gene_location/AGS_information/exon-intron_info/")
	   (psetf (start (car (exon-list *current-ags*)))
		  (parse-integer (get-attr-value ':|e_start| attributes))
		  (stop (car (exon-list *current-ags*)))
		  (parse-integer (get-attr-value ':|e_stop| attributes))))
	  ((xml-elt-matches-p "intron" "http://www.genomethreader.org/GTH_output/PGL_module/predicted_gene_location/AGS_information/exon-intron_info/")
	   (push (make-instance 'ags-intron-features
				:donor-prob (read-from-string
					     (get-attr-value ':|don_prob| attributes))
				:acceptor-prob (read-from-string
						(get-attr-value ':|acc_prob| attributes)))
		 (intron-list *current-ags*)))
	  ((xml-elt-matches-p "PGS_line" "http://www.genomethreader.org/GTH_output/PGL_module/predicted_gene_location/AGS_information/supporting_evidence/")
	   (setf *tmp-supp-pgs-alist* '()))
	  ((xml-elt-matches-p "exon" "http://www.genomethreader.org/GTH_output/PGL_module/predicted_gene_location/AGS_information/supporting_evidence/")
	   (push (cons (get-attr-value ':|start| attributes)
		       (get-attr-value ':|stop| attributes))
		 *tmp-supp-pgs-alist*))
	  ((or (xml-elt-matches-p "referenceDNA" "http://www.genomethreader.org/GTH_output/PGL_module/predicted_gene_location/AGS_information/supporting_evidence/")
	       (xml-elt-matches-p "referenceProtein" "http://www.genomethreader.org/GTH_output/PGL_module/predicted_gene_location/AGS_information/supporting_evidence/"))
	   ;; Assume that a particular query sequence contributes
	   ;; only one PGL_line to the alignment support of an AGS.
	   (setf (gethash (get-attr-value ':|id| attributes)
			  (support-hash *current-ags*))
		 (reverse *tmp-supp-pgs-alist*)
		 *tmp-supp-pgs-alist*
		 '()))
	  ((xml-elt-matches-p "gDNA" "http://www.genomethreader.org/GTH_output/PGL_module/predicted_gene_location/AGS_information/three_phase_translation/probable_ORFs/")
	   (if (string= "" (gdna-id *current-ags*))
	       (setf (gdna-id *current-ags*)
		     (get-attr-value ':|id| attributes))))
	  ((xml-elt-matches-p "none" "http://www.genomethreader.org/GTH_output/PGL_module/predicted_gene_location/AGS_information/three_phase_translation/probable_ORFs/")
	   (if (string= "" (gdna-id *current-ags*))
	       (setf (gdna-id *current-ags*)
		     (get-attr-value ':|gDNA_id| attributes))))
	  (t nil)))
  seed)

(defun parse-ags-gthxml-finish-element-hook (name attributes parent-seed seed)
  (declare (ignore attributes parent-seed))
  (macrolet ((xml-elt-matches-p (query-name namespace)
	       `(and (funcall *string-comparator* ,query-name name)
		 (funcall *string-comparator* ,namespace (get-current-xmlns)))))
    (cond ((xml-elt-matches-p "exon-intron_info" "http://www.genomethreader.org/GTH_output/PGL_module/predicted_gene_location/AGS_information/exon-intron_info/")
	   ;; Because exons and introns are cons'ed onto their respective lists,
	   ;; these lists must be reversed to preserve order w.r.t. to that in
	   ;; which they were encountered in the gthXML source stream.
	   (psetf (exon-list *current-ags*)
		  (nreverse (exon-list *current-ags*))
		  (intron-list *current-ags*)
		  (nreverse (intron-list *current-ags*))))
	  ((xml-elt-matches-p "AGS_information" "http://www.genomethreader.org/GTH_output/PGL_module/")
	   (push *current-ags* *agss-at-pgl*))
	  ((xml-elt-matches-p "predicted_gene_location" "http://www.genomethreader.org/GTH_output/PGL_module/")
	   (record-pgl-if (reverse *agss-at-pgl*)
			  :filter *op-specific-filter*
			  :test *op-specific-predicate*))
	  (t nil)))
  seed)

(defun parse-ags-gthxml (in-stream filter-for-operation predicate-for-operation)
  "After the dynamic environment for parsing AGS data from gthXML is established,
this function will set the hooks and invoke S-XML's start-parse-xml expression."
  (psetf *op-specific-filter* filter-for-operation
	 *op-specific-predicate* predicate-for-operation)
  (start-parse-xml in-stream
		   (make-instance 'xml-parser-state
				  :new-element-hook #'parse-ags-gthxml-new-element-hook
				  :finish-element-hook #'parse-ags-gthxml-finish-element-hook))
  (with-output-to-string (accumulator)
    (print-recorded-pgls (reverse *recorded-pgls*) accumulator)))

(defun find-all-ags (in-stream)
  "Parse AGSs from gthXML data, with no particular stringencies to speak of."
  (parse-ags-gthxml in-stream *identity-filter* *all-shall-pass*))

(defun find-altsplc-from-ags (in-stream)
  "Given AGS data from gthXML documents, identify alternative splicing events."
  ;; Because the processing of a document container will
  ;; require multiple invocations of this function,
  ;; validation of the *qual-aln-parms* and
  ;; *ags-alt-splc-parms* parameter objects will
  ;; be done only once, in the calling expression.
  (setf *query-alns-hash*
	(cadr (multiple-value-list (parse-qualified-alns-from-gthxml in-stream))))
  ;; After accumulating alignment info, we need to rewind the input stream
  ;; in order to process the alternative gene structure info.
  (if (not (file-position in-stream 0))
      (error "Unable to rewind input stream.")
      (parse-ags-gthxml in-stream 'retain-only-supported-agss 'alt-splc-detected-p)))
