;;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;; gthxml-aln-parse.lisp

;;;; Expressions to parse (and record) transcript and protein
;;;; spliced alignment features encountered in gthXML documents.

;;;; Michael E Sparks (mespar1@iastate.edu)
;;;; Last Modified: 5 February 2007

;;;; Copyright (c) 2007 Michael E Sparks
;;;; All rights reserved.
;;;;
;;;; You are granted the rights to distribute and use this software
;;;; as governed by the terms of the Lisp Lesser General Public License
;;;; (http://opensource.franz.com/preamble.html), also known as the LLGPL.

(in-package :toolsaq)

;;; Define various objects and their specialized methods:

(defclass query-aln-features ()
  ((query-id
    :accessor query-id
    :type string
    :documentation "Aligned transcript/protein identifier.")
   (query-is-transcript-p
    :accessor query-is-transcript-p
    :documentation "Is query of type cDNA (t) or Protein (nil)?")
   (query-orient
    :accessor query-orient
    :type character
    :documentation "Orientation of aligned transcript/protein.")
   (gdna-id
    :accessor gdna-id
    :type string
    :documentation "Genomic template identifier.")
   (gdna-orient
    :accessor gdna-orient
    :type character
    :documentation "Orientation of genomic template.")
   (total-aln-score
    :accessor total-aln-score
    :type short-float
    :documentation "Score of aligned transcript/protein.")
   (cum-len-scored-exons
    :accessor cum-len-scored-exons
    :type integer
    :documentation "Cumulative length of scored exons in alignment.")
   (percent-coverage
    :accessor percent-coverage
    :type short-float
    :documentation "Percent coverage of gDNA/cDNA/Protein in alignment.")
   (percent-coverage-type
    :accessor percent-coverage-type
    :type character
    :documentation "States whether percent coverage was given
in terms of G (gDNA), C (cDNA), or P (protein).")
   (gdna-exon-coors
    :accessor gdna-exon-coors
    :type list
    :initform '()
    :documentation "Stores exon start/stop
coordinates as integers in an alist.")
   (hqgdna-exon-coors
    ;; This field is sometimes irrelevant, e.g., quality adjusted
    ;; alignments aren't provided in GenomeThreader output.
    :accessor hqgdna-exon-coors
    :type list
    :initform '()
    :documentation "Stores quality adjusted exon start/stop
coordinates as integers in an alist.")
   (hq-aln-exists-p
    :accessor hq-aln-exists-p
    :initform nil
    :documentation "Was quality adjusted spliced alignment data encountered?"))
  (:documentation "Stores relevant features of an aligned transcript or protein (query)."))

(defmethod print-object ((object query-aln-features) stream)
  (print-unreadable-object (object stream :type nil :identity nil)
    (with-accessors ((gi gdna-id)
		     (go gdna-orient)
		     (qi query-id)
		     (qitp query-is-transcript-p)
		     (qo query-orient)
		     (tas total-aln-score)
		     (clse cum-len-scored-exons)
		     (pc percent-coverage)
		     (pct percent-coverage-type)
		     (gec gdna-exon-coors)
		     (hec hqgdna-exon-coors)
		     (haep hq-aln-exists-p))
      object
      (format stream "~&Query ID: ~a of type ~a, aligning in ~a orientation~%"
	      qi (if qitp "cDNA" "Protein") (case qo (#\+ "forward") (#\- "reverse")))
      (format stream "gDNA ID: ~a aligning in ~a orientation~%"
	      gi (case go (#\+ "forward") (#\- "reverse")))
      (format stream "Alignment Score: ~,3f~cLength: ~:dbp~cCoverage: ~,3f% of the ~a~%"
	      tas #\tab clse #\tab pc
	      (cond ((funcall *character-comparator* pct #\C) "cDNA")
		    ((funcall *character-comparator* pct #\P) "Protein")
		    (t "Genomic sequence")))
      (labels ((report-exon-coors (alist loc-string)
		 (if (endp alist)
		     loc-string
		     (report-exon-coors
		      (cdr alist) (concatenate 'string
					       loc-string " (" (caar alist)
					       " " (cdar alist) ")")))))
	(format stream "PGS exon coordinates (start stop):~a~%"
		(report-exon-coors gec ""))
	(when haep
	  (format stream  "Quality-adjusted PGS exon coordinates (start stop):~a~%"
		  (report-exon-coors hec "")))))))

(defvar *current-query* nil
  "Dynamic variable that records current query sequence,
i.e., a query-aln-features object.")

;;; Various user-adjustable parameters for governing aln-worthy-p's response.
;;; These are set to my own preferences, and can be adjusted by the user
;;; either at the REPL (preferred) or by modifying the sources directly.
;;; These will be described in the ToolSAQ API

(defclass qual-aln-parms ()
  ((cdna-of-interest-p
    :accessor cdna-of-interest-p
    :initform t
    :documentation "Are we interested in recording transcript (cDNA) alignments?")
   (protein-of-interest-p
    :accessor protein-of-interest-p
    :initform t
    :documentation "Are we interested in recording protein alignments?")
   (min-cdna-aln-score
    :accessor min-cdna-aln-score
    :type short-float
    :initform 0.8
    :documentation "Minimum alignment score of transcript (cDNA) alignment.")
   (min-protein-aln-score
    :accessor min-protein-aln-score
    :type short-float
    :initform 0.3
    :documentation "Minimum alignment score of protein alignment.")
   (min-tot-len-scored-exons
    :accessor min-tot-len-scored-exons
    :type integer
    :initform 100
    :documentation "Minimum length of all scored exons in alignment.")
   (min-percent-coverage
    :accessor min-percent-coverage
    :type short-float
    :initform 0.80
    :documentation "Minimum percent coverage of the alignment.")
   (force-high-type-be-query-type-p
    :accessor force-high-type-be-query-type-p
    :initform t
    :documentation "If query is cDNA, only consider alignments whose
percent-coverage-type slot is 'C'?  Same for protein queries, but with
said slot set to 'P'.")
   (force-hqpgs-be-present-p
    :accessor force-hqpgs-be-present-p
    :initform nil
    :documentation "Only consider alignments with quality adjusted results?"))
  (:documentation "Parameters governing identification of interesting alignments."))

(defmethod print-object ((object qual-aln-parms) stream)
  (print-unreadable-object (object stream :type nil :identity nil)
    (with-accessors ((coip cdna-of-interest-p)
		     (poip protein-of-interest-p)
		     (mcas min-cdna-aln-score)
		     (mpas min-protein-aln-score)
		     (mtlse min-tot-len-scored-exons)
		     (mpc min-percent-coverage)
		     (fhtbqtp force-high-type-be-query-type-p)
		     (fhbpp force-hqpgs-be-present-p))
      object
      (format stream "~&~a query types are being considered.~%"
	      (cond ((and coip poip) "cDNA and Protein")
		    (coip "Only cDNA")
		    (t "Only Protein")))
      (if coip (format stream "Min overall cDNA alignment score: ~,3f~%" mcas))
      (if poip (format stream "Min overall Protein alignment score: ~,3f~%" mpas))
      (format stream "Min total length of scored exons: ~d~%" mtlse)
      (format stream "Min percent coverage of aligned sequences: ~,3f" mpc)
      (format stream "~c(which ~a)~%"
	      #\tab
	      (if fhtbqtp
		  "must be of the query type"
		  "can be either of the query and genomic template"))
      (if fhbpp
	  (format stream "Requiring that quality adjusted PGS data be present.~%")
	  (format stream "Quality adjusted PGS data need not be present.~%")))))

(defmethod validate-parms-obj-qual ((obj qual-aln-parms))
  (with-accessors ((cintp cdna-of-interest-p)
		   (pintp protein-of-interest-p)
		   (mcsco min-cdna-aln-score)
		   (mpsco min-protein-aln-score)
		   (mlene min-tot-len-scored-exons)
		   (mpcov min-percent-coverage))
    obj
    (when (not (or cintp pintp))
      (cerror "Consider one, or possibly both."
	      "Ignoring both cDNA and Protein queries.")
      (loop (when (or cintp pintp) (return))
	    (psetf cintp (y-or-n-p "cDNA's of interest?")
		   pintp (y-or-n-p "Proteins of interest?"))))
    (if (and cintp (not (<= 0.0 mcsco 1.0)))
	(inquire-of-probability mcsco "minimum cDNA alignment score"))
    (if (and pintp (not (<= 0.0 mpsco 1.0)))
	(inquire-of-probability mpsco "minimum protein alignment score"))
    (if (minusp mlene)
	(fix-negative-value mlene "minimum length of scored exons"))
    (if (not (<= 0.0 mpcov 1.0))
	(inquire-of-probability mpcov "minimum percent coverage"))))

(defparameter *qual-aln-parms* (make-instance 'qual-aln-parms)
  "This object stores parameters influencing the identification
of qualifying cDNA and/or Protein alignments at genetic loci.")

;;; Code for actually recording interesting alignments

(defun aln-worthy-p (aln-feat-obj)
  "In considering aln-feat-obj, is its represented alignment of sufficient quality?"
  (with-accessors ((coip cdna-of-interest-p)
		   (poip protein-of-interest-p)
		   (mcas min-cdna-aln-score)
		   (mpas min-protein-aln-score)
		   (mtlse min-tot-len-scored-exons)
		   (mpc min-percent-coverage)
		   (fhtbqtp force-high-type-be-query-type-p))
    *qual-aln-parms*
    (with-accessors ((qitp query-is-transcript-p)
		     (tas total-aln-score)
		     (clse cum-len-scored-exons)
		     (pc percent-coverage)
		     (pct percent-coverage-type)
		     (hq-aln-exists-p hq-aln-exists-p))
      aln-feat-obj
      (and (or (and qitp coip)
	       (and (not qitp) poip))
	   (or (and qitp (>= tas mcas))
	       (and (not qitp) (>= tas mpas)))
	   (>= clse mtlse)
	   (>= pc mpc)
	   (if fhtbqtp
	       (or (and qitp (funcall *character-comparator* pct #\C))
		   (and (not qitp) (funcall *character-comparator* pct #\P)))
	       t)))))

(defvar *recorded-alns* nil
  "Hashtable of alignment features:

Keys   = string identifier of transcript/protein sequence
Values = list of locus-specific alignment features.")

(defun record-aln-if (aln-feat-obj &key (test *all-shall-pass*))
  "If the alignment is worth recording, do so."
  (when (funcall test aln-feat-obj)
    ;; Homologous sequences can multiply align to a given
    ;; genomic template, and there could be multiple such
    ;; templates processed in a given gthXML document.
    ;; So, we record alignments in a list.
    (macrolet ((hash-slot (afo) `(gethash (query-id ,afo) *recorded-alns*)))
      (unless (cadr (multiple-value-list (hash-slot aln-feat-obj)))
	(setf (hash-slot aln-feat-obj) '()))
      (push aln-feat-obj (hash-slot aln-feat-obj)))))

;;; Define and set hooks for the alignment parser

(defclass gthxml-aln-parse-state ()
  ((in-total-aln-score
    :initform nil
    :reader in-total-aln-score-p
    :writer (setf in-total-aln-score)
    :documentation "In the total alignment score state?")
   (in-cum-len-scored-exons
    :initform nil
    :reader in-cum-len-scored-exons-p
    :writer (setf in-cum-len-scored-exons)
    :documentation "In scored exon cumulative length state?"))
  (:documentation "Encapsulates bools tracking events during the gthXML parse."))

(defvar *event-trigger-status* (make-instance 'gthxml-aln-parse-state)
  "Records parser state w.r.t. gthXML features of interest;
used for invoking operations in the text hook.")

(defun parse-gthxml-aln-new-element-hook (name attributes seed)
  (macrolet ((xml-elt-matches-p (query-name namespace)
	       `(and (funcall *string-comparator* ,query-name name)
		 (funcall *string-comparator* ,namespace (get-current-xmlns)))))
    (cond ((xml-elt-matches-p "GTH_output" "http://www.genomethreader.org/GTH_output/")
	   ;; Note that it is assumed all strings denoting a
	   ;; particular cDNA or protein sequence in a gthXML
	   ;; document are case-identical.
	   (setf *recorded-alns* (make-hash-table :test #'equal)))
	  ((xml-elt-matches-p "spliced_alignment" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (setf *current-query* (make-instance 'query-aln-features)))
	  ((xml-elt-matches-p "total_alignment_score" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (setf (in-total-aln-score *event-trigger-status*) t))
	  ((xml-elt-matches-p "cumulative_length_of_scored_exons" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (setf (in-cum-len-scored-exons *event-trigger-status*) t))
	  ((xml-elt-matches-p "coverage" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (psetf (percent-coverage *current-query*)
		  (read-from-string (get-attr-value ':|percentage| attributes))
		  (percent-coverage-type *current-query*)
		  (character (get-attr-value ':|high_type| attributes))))
	  ((xml-elt-matches-p "gDNA" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (psetf (gdna-id *current-query*)
		  (get-attr-value ':|gen_id| attributes)
		  (gdna-orient *current-query*)
		  (character (get-attr-value ':|gen_strand| attributes))))
	  ((xml-elt-matches-p "rDNA" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (psetf (query-id *current-query*)
		  (get-attr-value ':|rDNA_id| attributes)
		  (query-is-transcript-p *current-query*)
		  t
		  (query-orient *current-query*)
		  (character (get-attr-value ':|rDNA_strand| attributes))))
	  ((xml-elt-matches-p "rProt" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (psetf (query-id *current-query*)
		  (get-attr-value ':|rProt_id| attributes)
		  (query-is-transcript-p *current-query*)
		  nil
		  ;; Proteins are always aligned in a forward sense
		  (query-orient *current-query*)
		  #\+))
	  ;; Though two nodes occur in the same namespace with the name "exon"
	  ;; they can be distinguished by presence of the "e_serial" attribute:
	  ((and (xml-elt-matches-p "exon" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
		(not (get-attr-value ':|e_serial| attributes)))
	   (push (cons (get-attr-value ':|e_start| attributes)
		       (get-attr-value ':|e_stop| attributes))
		 (gdna-exon-coors *current-query*)))
	  ((xml-elt-matches-p "hqPGS_line" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (setf (hq-aln-exists-p *current-query*) t))
	  ((xml-elt-matches-p "hqexon" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (push (cons (get-attr-value ':|e_start| attributes)
		       (get-attr-value ':|e_stop| attributes))
		 (hqgdna-exon-coors *current-query*)))
	  (t nil)))
  seed)

(defun parse-gthxml-aln-finish-element-hook (name attributes parent-seed seed)
  (declare (ignore attributes parent-seed))
  (macrolet ((xml-elt-matches-p (query-name namespace)
	       `(and (funcall *string-comparator* ,query-name name)
		 (funcall *string-comparator* ,namespace (get-current-xmlns)))))
    (cond ((xml-elt-matches-p "total_alignment_score" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (setf (in-total-aln-score *event-trigger-status*) nil))
	  ((xml-elt-matches-p "cumulative_length_of_scored_exons" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (setf (in-cum-len-scored-exons *event-trigger-status*) nil))
	  ((xml-elt-matches-p "gDNA_exon_coordinates" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (setf (gdna-exon-coors *current-query*)
		 (nreverse (gdna-exon-coors *current-query*))))
	  ((xml-elt-matches-p "hqgDNA_exon_coordinates" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (setf (hqgdna-exon-coors *current-query*)
		 (nreverse (hqgdna-exon-coors *current-query*))))
	  ((xml-elt-matches-p "spliced_alignment" "http://www.genomethreader.org/GTH_output/alignment_module/spliced_alignment/")
	   (record-aln-if *current-query* :test *op-specific-predicate*))
	  (t nil)))
  seed)

(defun parse-gthxml-aln-text-hook (string seed)
  (cond ((in-total-aln-score-p *event-trigger-status*)
	 (setf (total-aln-score *current-query*)
	       (read-from-string string)))
	((in-cum-len-scored-exons-p *event-trigger-status*)
	 (setf (cum-len-scored-exons *current-query*)
	       (parse-integer string)))
	(t nil))
  seed)

(defun parse-alns-gthxml (in-stream predicate-for-operation)
  "After the dynamic environment for parsing PGS and hqPGS data from
gthXML is established, this function will set the hooks and invoke
S-XML's start-parse-xml expression."
  (setf *op-specific-predicate* predicate-for-operation)
  (start-parse-xml in-stream
		   (make-instance 'xml-parser-state
				  :new-element-hook #'parse-gthxml-aln-new-element-hook
				  :finish-element-hook #'parse-gthxml-aln-finish-element-hook
				  :text-hook #'parse-gthxml-aln-text-hook))
  *recorded-alns*)

(defmacro parse-aln-form (stream fct)
  "Wraps the general form of alignment parsing functions."
  `(let ((local-hash (parse-alns-gthxml ,stream ,fct)))
    (values
     #'(lambda (stream)
	 (maphash #'(lambda (id alns)
		      (format stream
			      "~&~%*** Reporting alignments for query sequence ~a ***~%" id)
		      ;; print alignments in the order they
		      ;; were encountered in the gthXML stream
		      (let ((alns-size (list-length alns))
			    (aln-ind 0))
			(dolist (aln (reverse alns))
			  (format stream "Alignment ~a of ~a:~%~a~%"
				  (incf aln-ind) alns-size aln))))
		  local-hash))
     local-hash)))

(defun parse-qualified-alns-from-gthxml (in-stream)
  "Parse qualified PGS data from gthXML documents, returning two values:
first a closure for printing the hash to some stream,
and second, the hash itself."
  ;; Because the processing of a document container will
  ;; require multiple invocations of this function,
  ;; validation of the *qual-aln-parms* will
  ;; occur only once, in the calling expression.
  (parse-aln-form in-stream 'aln-worthy-p))

(defun parse-all-alns-from-gthxml (in-stream)
  "Parse all PGS data from gthXML documents, returning two values:
first a closure for printing the hash to some stream,
and second, the hash itself."
  (parse-aln-form in-stream *all-shall-pass*))
