;;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;; gthxml-parse-util.lisp

;;;; Expressions of general utility to gthXML parsing operations.

;;;; Michael E Sparks (mespar1@iastate.edu)
;;;; Last Modified: 3 February 2007

;;;; Copyright (c) 2007 Michael E Sparks
;;;; All rights reserved.
;;;;
;;;; You are granted the rights to distribute and use this software
;;;; as governed by the terms of the Lisp Lesser General Public License
;;;; (http://opensource.franz.com/preamble.html), also known as the LLGPL.

(in-package :toolsaq)

;;; Default predicates

(defconstant *all-shall-pass* #'(lambda (x) (declare (ignore x)) t)
  "Takes one argument, returns t.")

(defconstant *none-shall-pass* #'(lambda (x) (declare (ignore x)) nil)
  "Takes one argument, returns nil.") ; with apologies to Monty Python

(defvar *op-specific-predicate* *none-shall-pass*
  "Stores operation-specific function to use as predicate.")

;;; Default filters

(defconstant *identity-filter* #'(lambda (x) x)
  "Filter that permits all filtrate to proceed.")

(defconstant *trapping-filter* *none-shall-pass*
  "Filter that permits no filtrate to proceed.")

(defvar *op-specific-filter* *trapping-filter*
  "Binds a function for pre-filtering lists.")

;;; Condition definitions relevant to parameter integrity checking.

(define-condition bad-parameter-value (error)
  ((culprit
    :initarg :culprit
    :reader culprit
    :documentation "The parameter of suspect value.")
   (value
    :initarg :value
    :reader value
    :documentation "Suspect value of the parameter.")
   (culprit-type
    :initarg :culprit-type
    :reader culprit-type
    :documentation "Type of the parameter of suspect value."))
  (:documentation "Condition for parameters having suspect values."))

(define-condition invalid-real (bad-parameter-value) ()
  (:report (lambda (condition stream)
	     (format stream "The ~A (= ~A) is not a ~A."
		     (culprit condition)
		     (value condition)
		     (culprit-type condition))))
  (:documentation "Condition for invalid probabilities."))

(define-condition negative-value (bad-parameter-value) ()
  (:report (lambda (condition stream)
	     (format stream "The ~A (= ~A) must be non-negative."
		     (culprit condition) (value condition))))
  (:documentation "Condition for negative-valued variables that should not be."))

;;; Generic function for testing parameter objects,
;;; and possibly raising the above conditions.

(defgeneric validate-parms-obj-qual (obj)
  (:documentation "Method to assure integrity of parameter objects."))

;;; Expressions for handling raised conditions

(defun prompt-for ()
  "Queries the user for a value, then returns a list of it."
  (format *query-io* "~&Enter a new value: ")
  (list (eval (read *query-io*))))

(defmacro fix-negative-value (var var-desc)
  "Fixes parameters that invoke a negative-value error."
  `(loop (when (not (minusp (eval ,var))) (return ,var))
    (restart-case (error 'negative-value
			 :culprit ,var-desc
			 :value (eval ,var))
      (store-value (val)
	:report (lambda (stream)
		  (format stream "Replace value."))
	:interactive prompt-for
	(setf ,var val)))))

(defmacro fix-bad-ratio (var var-desc)
  "Fixes ratios that invoke an invalid-real error."
  `(loop (when (not (minusp (eval ,var))) (return ,var))
    (restart-case (error 'invalid-real
			 :culprit ,var-desc
			 :value (eval ,var)
			 :culprit-type "reasonable some:all ratio")
      (store-value (val)
	:report (lambda (stream)
		  (format stream "Replace value."))
	:interactive prompt-for
	(setf ,var val)))))

(defmacro inquire-of-probability (var var-desc)
  "Corrects probabilities that invoke an invalid-real error, if desired."
  `(restart-case (error 'invalid-real
		  :culprit ,var-desc
		  :value (eval ,var)
		  :culprit-type "valid probability")
    (continue ()
     :report "Use value anyway.")
    (store-value (val)
     :report (lambda (stream)
	       (format stream "Replace value."))
     :interactive prompt-for
     (setf ,var val))))
